# Writing an HTTP Server in Cython+

[Cython+](https://www.cython.plus) is an ongoing project to extend [Cython](https://cython.org/) with [high performance GIL-free types](https://www.youtube.com/watch?v=Lb6q1hhBp-k) using static typing.

Our [next steps](https://www.cython.plus/P-CYP-Roadmap) include with writing a minimal standard library.

To motivate this effort I am writing an HTTP Server in Cython+, building everything I will need along the way.


## Usage

The repository looks like this:

```
- README.md
- src/
    - Makefile
    - setup.py
    - stdlib/
        - *.pxd
        - ...
    - *.pyx
    - ...
- ...
```

Each `.pyx` file in `src` is a standalone demo.

To run the demos (e.g. `helloworld.pyx`):

```
cd src
make helloworld
```


## GIL-free strings

Until now Cython+ merely used C++'s `std::string` to represent strings.

However I realised this would not be enough to accurately wrap the C socket APIs: they often accept `null` as a valid and meaningful value for a `char *` string argument. But there is no easy way to represent `null` with `std::string`.

So instead I wrote a minimal string implementation by wrapping a `std::string_view` (C++17) into a Cython+ GIL-free type. This is one of our next steps anyway. There are multiple advantages:

- full control over memory management: unlike `std::string`, `std::string_view` doesn't manage memory.
- in particular, it will let us take ownership of `char *` strings allocated by a C API without copying.
- aliasing a string will not create a copy, unlike with `std::string`.
- we can implement our own string methods.
- in particular, we can enforce immutability.

And since Cython+ GIL-free types are allocated to the heap and referenced by pointer, `null` is a valid value.

This is still an unfinished implementation, so I'll add to it as I need.


## Echo Server and Client

Using [Beej's Guide to Network Programming](https://beej.us/guide/bgnet/html//index.html) as a handy reference, I wrote adapters for needed parts of the POSIX socket API. As much as practical, these adapters mirror [Python's socket API](https://docs.python.org/3/library/socket.html).

These adapters include the functions:
- `getaddrinfo`
- `socket`
- `connect`
- `bind`
- `listen`
- `accept`
- `recv`
- `send`
- `shutdown`
- `close`

The new GIL-free strings proved very useful for this :)

Also, since currently Cython+ doesn't have GIL-free exceptions (also in our next steps), I take the GIL and raise Python exceptions instead when there is an error.

It's not complete, but it's enough to write a minimal echo server and client demo:

```
make server
```

and

```
make client
```

